/*
 * Port to gx for wii by Jim Rees November 2008
 * Some code cribbed from nehe lesson 3 port to GX by WinterMute
 *
 * cbzone_gpr.c
 *  -- Todd W Mummert, December 1990, CMU
 *
 * RCS info
 *  $Header: c_gpr.c,v 1.1 91/01/12 02:03:32 mummert Locked $
 *
 * emulate gpr/ftn on top of X11.
 *
 * I don't know who originally wrote this emulation routine, but
 * I've seriously changed it anyway.  Therefore, it should not
 * be used with any expectation that it will do anything similiar
 * to the original gpr routines.
 * Almost all of the routines have had their arguments shortened.
 * And why bother returning status when it's always the same.  Most
 * importantly I removed the mallocs that were in polyline and
 * multiline.
 *
 * Added support for window managers, iconification, color, etc...
 * Now, it's basically a file full of one line functions...except
 * for the event handling and initialization.
 *
 * let's define how we are going to use the GC's...
 * Erase GC...both foreground and background the color of the game
 *   background...
 * Text GC...foreground is red...as all text is currently
 * Draw GC...will be changed often...usually just the color
 * BitBlt GC...it will handle the printing of the
 *           lander/missile warning boxes after they are initially
 *           placed.
 */

#include "c_includes.h"

#include "bitmaps/moon.bits"
#include "bitmaps/joystick.bits"
#include "bitmaps/hswitch.bits"
#include "bitmaps/lswitch.bits"
#include "bitmaps/tank.bits"
#include "bitmaps/cursor.bits"
#include "bitmaps/cshape.bits"
#include "rowmans.c"

#define PAD_INC 8
#define DEFAULT_FIFO_SIZE	(256*1024)

/*
 * cbzone scale seems to be about 0,0 : 1000,660, with origin in upper left
 * Wii/gx seems to be -1,-1 : 1,1
 * This could probably be scaled in the gx library but I'm going to do it here
 */

// #define NEW
#ifdef NEW
#define spx(x) ((float) (x) / 3.)
#define spy(y) (639. - (float) (y) / 3.)
#else
#define spx(x) ((float) (x) / 500. - 1.)
#define spy(y) (1. - (float) (y) / 500.)
#endif

#define xsp(x) ((int) (((x) + 1.) * 500.))
#define ysp(y) ((int) ((1. - (y)) * 500.))

typedef struct {
    char name[14];		/* the color will be stored as  */
} Colorname;			/*      #RRRRGGGGBBBB           */

typedef struct {
    int width;
    int height;
    char *bits;
    unsigned int color;
} Bmap;

Bool paused = False;

void *frameBuffer[2] = { NULL, NULL };

u32 fb;
GXRModeObj *rmode;
Mtx view, GXmodelView2D;

/* Text and drawing colors */
GXColor tcolor, tbgcolor, dcolor;

/* Cursor position */
Position_t mouse_posn;

#define NUMPIXMAPS 5
Bmap bmaps[NUMPIXMAPS] = {
    { moon_width, moon_height, moon_bits, COLOR_MOON },
    { joystick_width, joystick_height, joystick_bits, COLOR_JOYSTICK },
    { hswitch_width, hswitch_height, hswitch_bits, COLOR_TEXT },
    { lswitch_width, lswitch_height, lswitch_bits, COLOR_TEXT },
    { tank_width, tank_height, tank_bits, COLOR_TEXT },
};

struct {
    int cidx;
    char *cname;
    unsigned int cvalue;
} colors[] = {
    { COLOR_BG, "black", 0x000000 },
    { COLOR_FG, "skyblue", 0x87ceeb },
    { COLOR_TEXT, "red", 0xff0000 },
    { COLOR_ESALVO, "red", 0xff0000 },
    { COLOR_PSALVO, "skyblue", 0x87ceeb },
    { COLOR_JOYSTICK, "red", 0xff0000 },
    { COLOR_SCANNER, "green", 0x00ff00 },
    { COLOR_HORIZON, "gray50", 0x808080 },
    { COLOR_MOON, "gray50", 0x808080 },
    { COLOR_XHAIR, "gray50", 0x808080 },
    { COLOR_CURSOR, "skyblue", 0x87ceeb },
    { COLOR_CRACKS, "skyblue", 0x87ceeb },
    { COLOR_TANK, "green", 0x00ff00 },
    { COLOR_SUPER, "green", 0x00ff00 },
    { COLOR_MISSILE, "green", 0x00ff00 },
    { COLOR_COPTER, "green", 0x00ff00 },
    { COLOR_LANDER, "green", 0x00ff00 },
    { COLOR_CUBE, "yellow", 0xffff00 },
    { COLOR_PYRAMID, "yellow", 0xffff00 },
    { -1, NULL, 0xffffff }
};

void
grabpointer(void)
{
}

void
ungrabpointer(void)
{
}

void
gprinit(void)
{
    int i;
    static f32 w, h, yscale;
    static u32 xfbHeight;
    static Mtx perspective;
    static GXColor background = { 0, 0, 0, 0xff };
    static Vector cam = { 0.0F, 0.0F, 0.0F },
	up = { 0.0F, 1.0F, 0.0F},
	look = { 0.0F, 0.0F, -1.0F};
    void *gp_fifo;

    // init the vi.
    VIDEO_Init();
    WPAD_Init();

    rmode = VIDEO_GetPreferredMode(NULL);

    // allocate 2 framebuffers for double buffering
    frameBuffer[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
    frameBuffer[1] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));

#ifdef HAVE_PRINTF
    console_init(frameBuffer[fb], 20, 20, rmode->fbWidth, rmode->xfbHeight,
		 rmode->fbWidth * VI_DISPLAY_PIX_SZ);
#endif

    VIDEO_Configure(rmode);
    VIDEO_SetNextFramebuffer(frameBuffer[fb]);
    VIDEO_SetBlack(FALSE);
    VIDEO_Flush();
    VIDEO_WaitVSync();
    if (rmode->viTVMode & VI_NON_INTERLACE)
	VIDEO_WaitVSync();

    // setup the fifo and then init the flipper
    gp_fifo = memalign(32, DEFAULT_FIFO_SIZE);
    memset(gp_fifo, 0, DEFAULT_FIFO_SIZE);

    GX_Init(gp_fifo, DEFAULT_FIFO_SIZE);

    // clears the bg to color and clears the z buffer
    GX_SetCopyClear(background, 0x00ffffff);

    // other gx setup
    GX_SetViewport(0, 0, rmode->fbWidth, rmode->efbHeight, 0, 1);
    yscale = GX_GetYScaleFactor(rmode->efbHeight, rmode->xfbHeight);
    xfbHeight = GX_SetDispCopyYScale(yscale);
    GX_SetScissor(0, 0, rmode->fbWidth, rmode->efbHeight);
    GX_SetDispCopySrc(0, 0, rmode->fbWidth, rmode->efbHeight);
    GX_SetDispCopyDst(rmode->fbWidth, xfbHeight);
    GX_SetCopyFilter(rmode->aa, rmode->sample_pattern, GX_TRUE, rmode->vfilter);
    GX_SetFieldMode(rmode->field_rendering,
		    ((rmode->viHeight == 2 * rmode->xfbHeight) ? GX_ENABLE : GX_DISABLE));

    GX_SetCullMode(GX_CULL_NONE);
    GX_CopyDisp(frameBuffer[fb], GX_TRUE);
    GX_SetDispCopyGamma(GX_GM_1_0);

    // setup the vertex descriptor
    // tells the flipper to expect direct data
    GX_ClearVtxDesc();
    GX_SetVtxDesc(GX_VA_POS, GX_DIRECT);
    GX_SetVtxDesc(GX_VA_CLR0, GX_DIRECT);

    // setup the vertex attribute table
    // describes the data
    // args: vat location 0-7, type of data, data format, size, scale
    // so for ex. in the first call we are sending position data with
    // 3 values X,Y,Z of size F32. scale sets the number of fractional
    // bits for non float data.
    GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_POS, GX_POS_XYZ, GX_F32, 0);
    GX_SetVtxAttrFmt(GX_VTXFMT0, GX_VA_CLR0, GX_CLR_RGBA, GX_RGB8, 0);

    GX_SetNumChans(1);
    GX_SetNumTexGens(0);
    GX_SetTevOrder(GX_TEVSTAGE0, GX_TEXCOORDNULL, GX_TEXMAP_NULL, GX_COLOR0A0);
    GX_SetTevOp(GX_TEVSTAGE0, GX_PASSCLR);

#ifdef NEW
    guMtxIdentity(GXmodelView2D);
    guMtxTransApply (GXmodelView2D, GXmodelView2D, 0.0F, 0.0F, -50.0F);
    GX_LoadPosMtxImm(GXmodelView2D, GX_PNMTX0);

    guOrtho(perspective, 0, 479, 0, 639, 0, 300);
    GX_LoadProjectionMtx(perspective, GX_ORTHOGRAPHIC);
    GX_SetViewport(0, 0, rmode->fbWidth, rmode->efbHeight, 0, 1);
#else
    guLookAt(view, &cam, &up, &look);

    // setup our projection matrix
    // this creates a perspective matrix with a view angle of 90,
    // and aspect ratio based on the display resolution
    w = rmode->viWidth;
    h = rmode->viHeight;
    guPerspective(perspective, 45, (f32) w / h, 0.1F, 300.0F);
    GX_LoadProjectionMtx(perspective, GX_PERSPECTIVE);
#endif

    /* Original code used a color map, we don't */
    for (i = 0; i < MAX_COLORS; i++)
	opt->cpi[i] = i;

    mouse_posn.x = 500;
    mouse_posn.y = 355;
}

void
gprstart(void)
{
    static Mtx model, modelview;

    GX_SetViewport(0, 0, rmode->fbWidth, rmode->efbHeight, 0, 1);
    guMtxIdentity(model);
    guMtxTransApply(model, model, 0.f, 0.f, -4.f);
    guMtxConcat(view, model, modelview);
    GX_LoadPosMtxImm(modelview, GX_PNMTX0);
}

void
polyline_c(void *xp, int n, GXColor *cp)
{
    Coord2dp p = xp;
    int i;

    GX_Begin(GX_LINES, GX_VTXFMT0, (n - 1) * 2);

    for (i = 0; i < n - 1; i++) {
	GX_Position3f32(spx(p[i].x), spy(p[i].y), 0.0f);
	GX_Color3u8(cp->r, cp->g, cp->b);
	GX_Position3f32(spx(p[i + 1].x), spy(p[i + 1].y), 0.0f);
	GX_Color3u8(cp->r, cp->g, cp->b);
    }

    GX_End();
}

void
polyline(void *xp, int n)
{
    polyline_c(xp, n, &dcolor);
}

void
multiline(void *xseg, int n)
{
    XSegment *s = xseg;
    int i;
    GXColor *cp = &dcolor;

    GX_Begin(GX_LINES, GX_VTXFMT0, n * 2);

    for (i = 0; i < n; i++) {
	GX_Position3f32(spx(s[i].x1), spy(s[i].y1), 0.0f);
	GX_Color3u8(cp->r, cp->g, cp->b);
	GX_Position3f32(spx(s[i].x2), spy(s[i].y2), 0.0f);
	GX_Color3u8(cp->r, cp->g, cp->b);
    }

    GX_End();
}

void
drawrectangle(int x, int y, int width, int height)
{
    Coord2d c[5];

    c[0].x = x;
    c[0].y = y;
    c[1].x = x + width;
    c[1].y = y;
    c[2].x = x + width;
    c[2].y = y + height;
    c[3].x = x;
    c[3].y = y + height;
    c[4].x = x;
    c[4].y = y;

    polyline(c, 5);
}

void
gprdone(void)
{
    GX_DrawDone();
    fb ^= 1;			// flip framebuffer
    GX_SetZMode(GX_TRUE, GX_LEQUAL, GX_TRUE);
    GX_SetColorUpdate(GX_TRUE);
    GX_CopyDisp(frameBuffer[fb], GX_TRUE);

    VIDEO_SetNextFramebuffer(frameBuffer[fb]);
    VIDEO_Flush();
    VIDEO_WaitVSync();
}

void
bitblt(Window_t * window)
{
#if 0
    XFillRectangle(d, w, BitBltGC, window->base.x, window->base.y,
		   window->size.x, window->size.y);
#endif
}

void
clearrectangle(Window_t * window, Position_t * dsto)
{
    GX_Begin(GX_QUADS, GX_VTXFMT0, 4);

    GX_Position3f32(spx(dsto->x), spy(dsto->y), 0.0f);
    GX_Color3u8(0, 0, 0);
    GX_Position3f32(spx(dsto->x + window->size.x), spy(dsto->y), 0.0f);
    GX_Color3u8(0, 0, 0);
    GX_Position3f32(spx(dsto->x + window->size.x),
		    spy(dsto->y + window->size.y), 0.0f);
    GX_Color3u8(0, 0, 0);
    GX_Position3f32(spx(dsto->x), spy(dsto->y + window->size.y), 0.0f);
    GX_Color3u8(0, 0, 0);

    GX_End();
}

void
gprsetclippingactive(Bool flag)
{
}

void
tonetime(void)
{
}

void
gprinqcursor(int *p)
{
    *((Position_t *) p) = mouse_posn;
}

Bool
gprcondeventwait(char *key, Position_t *posn)
{
    Bool rv;
    static u32 p0 = 0, p1, pd, pu;

    WPAD_ScanPads();
    p1 = WPAD_ButtonsHeld(0);
    pd = p1 & (p0 ^ p1);
    pu = p0 & (p0 ^ p1);
    p0 = p1;

    rv = False;
    *key = '\0';

    /* This won't work if two buttons are pressed at once */

    if (pd & WPAD_BUTTON_HOME)
	*key = 'Q';
    else if (pd & WPAD_BUTTON_A)
	*key = 'a';
    else if (pu & WPAD_BUTTON_A)
	*key = 'A';
    else if (pd & WPAD_BUTTON_B)
	*key = 'b';
    else if (pu & WPAD_BUTTON_B)
	*key = 'B';

    if (p1 & WPAD_BUTTON_LEFT) {
	mouse_posn.x -= PAD_INC;
	if (mouse_posn.x < 0)
	    mouse_posn.x = 0;
    } else if (p1 & WPAD_BUTTON_RIGHT) {
	mouse_posn.x += PAD_INC;
	if (mouse_posn.x > 1000)
	    mouse_posn.x = 1000;
    }
    if (p1 & WPAD_BUTTON_UP) {
	mouse_posn.y -= PAD_INC;
	if (mouse_posn.y < 0)
	    mouse_posn.y = 0;
    } else if (p1 & WPAD_BUTTON_DOWN) {
	mouse_posn.y += PAD_INC;
	if (mouse_posn.y > 710)
	    mouse_posn.y = 710;
    }

    if (*key != '\0')
	rv = True;

    if (!paused)
	*posn = mouse_posn;

    return rv;
}

void
gprsetcursorposition(int *p)
{
    mouse_posn = *((Position_t *) p);
}

void
gprsetvalue(int idx, GXColor *cp)
{
    int i, v;

    for (i = 0; colors[i].cname != NULL; i++) {
	if (colors[i].cidx == idx)
	    break;
    }

    v = colors[i].cvalue;

    cp->r = (v >> 16) & 0xff;
    cp->g = (v >> 8) & 0xff;
    cp->b = v & 0xff;
}

void
gprsettextvalue(int idx)
{
    gprsetvalue(idx, &tcolor);
}

void
gprsettextbackgroundvalue(int idx)
{
    gprsetvalue(idx, &tbgcolor);
}

void
gprsetdrawvalue(int idx)
{
    gprsetvalue(idx, &dcolor);
}

void
gprcircle(int *center, int radius)
{
    Position_t *ctr = (Position_t *) center;
}

void
gprcirclefilled(int *center, int radius)
{
    Position_t *ctr = (Position_t *) center;
}

void
gprsetclipwindow(Window_t * window)
{
}

void
clearentirescreen(void)
{
    VIDEO_ClearFrameBuffer(rmode, frameBuffer[fb], COLOR_BLACK);
}

void
flushwindow(void)
{
    VIDEO_Flush();
}

void
waitforkey(char c)
{
    WPAD_ScanPads();
    u32 pressed = WPAD_ButtonsDown(0);
    if (pressed & WPAD_BUTTON_HOME)
	exit(0);
}

void
putpixmap(int i, int *p)
{
}

void
removepixmap(int i, int *p)
{
}

#ifndef HAVE_RANDOM
static long random_number = 1;

long
random(void)
{
    random_number = random_number * 1103515245 + 12345;
    return (random_number & 0x7fffffff);
}

void
srandom(unsigned int seed)
{
    random_number = seed;
}
#endif

#define lscale(x) ((x) * 3 / 5)

void
drawglyph(int n, int *xp, int *yp)
{
    int i, pi;
    struct strokes *s;
    struct font *f;
    Coord2d p[100];

    if (n > 95)
	return;

    f = &font[n];
    s = f->s;
    *xp -= lscale(f->l);

    for (i = pi = 0; i < f->glen; s++, i++) {
	if (!(s->pen) && pi > 0) {
	    polyline_c(p, pi, &tcolor);
	    pi = 0;
	}
	p[pi].x = lscale(s->x) + *xp;
	p[pi].y = lscale(s->y) + *yp;
	pi++;
    }

    if (pi > 0)
	polyline_c(p, pi, &tcolor);

    *xp += lscale(f->r);
}

void
printstring(int x, int y, char *s, int nchars)
{
    while (*s != '\0')
	drawglyph((int) *s++ - ' ', &x, &y);
}
