#include "c_includes.h"
/*
 * cbzone_parseopts.c
 *  -- Todd W Mummert, December 1990, CMU
 *
 * RCS Info
 *  $Header: c_parseopts.c,v 1.1 91/01/12 02:03:36 mummert Locked $
 *
 * Parse the options, read the MOTD, etc...
 *
 * The prints in these routines will never go the game window, as it
 * does not exist.  cbzone cannot be backgrounded from the beginning
 * since the execl will block on tty output.  To allow this we create
 * yet another flag that specifies that motd is not to be read.
 */

int
pager(char *file)
{
    return 0;
}

int
getoptionint(char *s)
{
    char rest[100];
    int num;

    if (sscanf(s, "%d%s", &num, rest) != 1) {
	printf("Error in optional argument %s; use -help for help.\n", s);
	exit(0);
    }
    return (num);
}

/*
 * the following routine may be called in one of two ways...
 *  either w/ the display set or without...if without, then we
 *  will need to parse all the options...otherwise the resources should
 *  have taken care of most of them for us.
 *
 *  now even if the display is set, we may get options that were
 *  ambiguous.
 *
 *  since options are more than one letter, we can't use getopt.
 */
#define MAXOPTIONS 16
#define OPTIONINT 7
void
parseopt(int argc, char *argv[], Bool status)
{
    opt->delay = DELAY;
    opt->mblocks = MBLOCKS;
    opt->mlanders = MLANDERS;
    opt->mtanks = MTANKS;
    opt->mmissiles = MMISSILES;
    opt->msalvos = MSALVOS;
    opt->copters = 0;
    opt->loud = 1;
    opt->scores = 0;
    opt->original = 0;
    opt->version = 0;
    opt->help = 0;
    opt->output = 1;
    opt->mono = 0;
    opt->cursor = 0;
    opt->fullscreen = 1;

    opt->menemies = (opt->mtanks > opt->mmissiles ?
		     opt->mtanks : opt->mmissiles);
    if (!opt->menemies) {
	printf("Must have at least one missile or tank.\n");
	exit(1);
    }

    if (opt->msalvos == -1)
	opt->msalvos = opt->menemies;
    opt->mobjects = opt->mblocks + opt->mlanders + 2*opt->menemies +
	opt->msalvos + 1;
    opt->estart = 1;
    opt->lstart = opt->estart + opt->menemies;
    opt->sstart = opt->lstart + opt->mlanders;
    opt->bstart = opt->sstart + opt->menemies + opt->msalvos;

    if (opt->mmissiles == MMISSILES &&
	opt->mtanks == MTANKS &&
	opt->mlanders == MLANDERS &&
	opt->mblocks == MBLOCKS &&
	opt->delay <= DELAY &&
	opt->msalvos == opt->menemies)
	opt->practice = False;
    else
	opt->practice = True;
}
